@@ -2,22 +2,29 @@
 #include "ev-job-queue.h"
 #include "ev-page-cache.h"
 #include "ev-selection.h"
+#include "ev-document-images.h"
+#include "ev-image.h"
 
 typedef struct _CacheJobInfo
 {
 	EvJob *job;
-	GdkPixbuf *pixbuf;
 	EvRenderContext *rc;
+
+	/* Data we get from rendering */
+	cairo_surface_t *surface;
 	GList *link_mapping;
+	GList *image_mapping;
 	GdkRegion *text_mapping;
 	
-	/* Selection info.  If the *_points structs are unset, we put -1 in x1.
-	 * selection_points are the coordinates encapsulated in selection.
-	 * new_points is the target selection size. */
+	/* Selection data. 
+	 * Selection_points are the coordinates encapsulated in selection.
+	 * target_points is the target selection size. */
 	EvRectangle selection_points;
-	GdkPixbuf *selection;
+	EvRectangle target_points;
+	gboolean    points_set;
+	
+	cairo_surface_t *selection;
 	GdkRegion *selection_region;
-	EvRectangle new_points;
 } CacheJobInfo;
 
 struct _EvPixbufCache
@@ -67,10 +74,10 @@ static CacheJobInfo *find_job_cache             (EvPixbufCache      *pixbuf_cach
 static void          copy_job_to_job_info       (EvJobRender        *job_render,
 						 CacheJobInfo       *job_info,
 						 EvPixbufCache      *pixbuf_cache);
-static gboolean      new_selection_pixbuf_needed(EvPixbufCache      *pixbuf_cache,
-						 CacheJobInfo       *job_info,
-						 gint                page,
-						 gfloat              scale);
+static gboolean      new_selection_surface_needed(EvPixbufCache      *pixbuf_cache,
+						  CacheJobInfo       *job_info,
+						  gint                page,
+						  gfloat              scale);
 
 
 /* These are used for iterating through the prev and next arrays */
@@ -124,6 +131,8 @@ ev_pixbuf_cache_finalize (GObject *object)
 	g_free (pixbuf_cache->prev_job);
 	g_free (pixbuf_cache->job_list);
 	g_free (pixbuf_cache->next_job);
+
+	G_OBJECT_CLASS (ev_pixbuf_cache_parent_class)->finalize (object);
 }
 
 static void
@@ -140,29 +149,36 @@ dispose_cache_job_info (CacheJobInfo *job_info,
 		g_object_unref (G_OBJECT (job_info->job));
 		job_info->job = NULL;
 	}
-	if (job_info->pixbuf) {
-		g_object_unref (G_OBJECT (job_info->pixbuf));
-		job_info->pixbuf = NULL;
+	if (job_info->surface) {
+		cairo_surface_destroy (job_info->surface);
+		job_info->surface = NULL;
 	}
 	if (job_info->link_mapping) {
 		ev_link_mapping_free (job_info->link_mapping);
 		job_info->link_mapping = NULL;
 	}
+	if (job_info->image_mapping) {
+		ev_image_mapping_free (job_info->image_mapping);
+		job_info->image_mapping = NULL;
+	}
 	if (job_info->text_mapping) {
 		gdk_region_destroy (job_info->text_mapping);
 		job_info->text_mapping = NULL;
 	}
 	if (job_info->selection) {
-		g_object_unref (G_OBJECT (job_info->selection));
+		cairo_surface_destroy (job_info->selection);
 		job_info->selection = NULL;
 	}
 	if (job_info->selection_region) {
 		gdk_region_destroy (job_info->selection_region);
 		job_info->selection_region = NULL;
 	}
+	if (job_info->rc) {
+		g_object_unref (G_OBJECT (job_info->rc));
+		job_info->rc = NULL;
+	}
 
-	job_info->selection_points.x1 = -1;
-	job_info->new_points.x1 = -1;
+	job_info->points_set = FALSE;
 }
 
 static void
@@ -181,6 +197,8 @@ ev_pixbuf_cache_dispose (GObject *object)
 	for (i = 0; i < PAGE_CACHE_LEN (pixbuf_cache); i++) {
 		dispose_cache_job_info (pixbuf_cache->job_list + i, pixbuf_cache);
 	}
+
+	G_OBJECT_CLASS (ev_pixbuf_cache_parent_class)->dispose (object);
 }
 
 
@@ -204,7 +222,6 @@ job_finished_cb (EvJob         *job,
 {
 	CacheJobInfo *job_info;
 	EvJobRender *job_render = EV_JOB_RENDER (job);
-	GdkPixbuf *pixbuf;
 
 	/* If the job is outside of our interest, we silently discard it */
 	if ((job_render->rc->page < (pixbuf_cache->start_page - pixbuf_cache->preload_cache_size)) ||
@@ -215,39 +232,7 @@ job_finished_cb (EvJob         *job,
 	
 	job_info = find_job_cache (pixbuf_cache, job_render->rc->page);
 
-	pixbuf = g_object_ref (job_render->pixbuf);
-	if (job_info->pixbuf)
-		g_object_unref (job_info->pixbuf);
-	job_info->pixbuf = pixbuf;
-
-	if (job_render->link_mapping) {
-		if (job_info->link_mapping)
-			ev_link_mapping_free (job_info->link_mapping);
-		job_info->link_mapping = job_render->link_mapping;
-	}
-
-	if (job_render->text_mapping) {
-		if (job_info->text_mapping)
-			gdk_region_destroy (job_info->text_mapping);
-		job_info->text_mapping = job_render->text_mapping;
-	}
-
-	if (job_render->include_selection) {
-		pixbuf = g_object_ref (job_render->selection);
-		if (job_info->selection)
-			g_object_unref (job_info->selection);
-		if (job_info->selection_region)
-			gdk_region_destroy (job_info->selection_region);
-		job_info->selection_points = job_render->selection_points;
-		job_info->selection_region = gdk_region_copy (job_render->selection_region);
-		job_info->selection = pixbuf;
-		g_assert (job_info->selection_points.x1 >= 0);
-	}
-	
-	if (job_info->job == job)
-		job_info->job = NULL;
-	g_object_unref (job);
-
+	copy_job_to_job_info (job_render, job_info, pixbuf_cache);
 	g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 }
 
@@ -255,7 +240,8 @@ job_finished_cb (EvJob         *job,
  * given a scale.  If it won't, it removes the job and clears it to NULL.
  */
 static void
-check_job_size_and_unref (CacheJobInfo *job_info,
+check_job_size_and_unref (EvPixbufCache *pixbuf_cache,
+			  CacheJobInfo *job_info,
 			  EvPageCache  *page_cache,
 			  gfloat        scale)
 {
@@ -277,12 +263,11 @@ check_job_size_and_unref (CacheJobInfo *job_info,
 	    height == EV_JOB_RENDER (job_info->job)->target_height)
 		return;
 
-	/* Try to remove the job.  If we can't, then the thread has already
-	 * picked it up and we are going get a signal when it's done.  If we
-	 * can, then the job is fully dead and will never rnu.. */
-	if (ev_job_queue_remove_job (job_info->job))
-		g_object_unref (job_info->job);
-
+	g_signal_handlers_disconnect_by_func (job_info->job,
+					      G_CALLBACK (job_finished_cb),
+					      pixbuf_cache);
+	ev_job_queue_remove_job (job_info->job);
+	g_object_unref (job_info->job);
 	job_info->job = NULL;
 }
 
@@ -335,8 +320,9 @@ move_one_job (CacheJobInfo  *job_info,
 
 	*target_page = *job_info;
 	job_info->job = NULL;
-	job_info->pixbuf = NULL;
+	job_info->surface = NULL;
 	job_info->link_mapping = NULL;
+	job_info->image_mapping = NULL;
 
 	if (new_priority != priority && target_page->job) {
 		ev_job_queue_update_job (target_page->job, new_priority);
@@ -421,17 +407,62 @@ copy_job_to_job_info (EvJobRender   *job_render,
 		      CacheJobInfo  *job_info,
 		      EvPixbufCache *pixbuf_cache)
 {
-	GdkPixbuf *pixbuf;
 
-	pixbuf = g_object_ref (job_render->pixbuf);
+	job_info->points_set = FALSE;
 
-	dispose_cache_job_info (job_info, pixbuf_cache);
+	if (job_info->surface) {
+		cairo_surface_destroy (job_info->surface);
+	}
+	job_info->surface = cairo_surface_reference (job_render->surface);
 
-	job_info->pixbuf = pixbuf;
-	if (job_render->link_mapping)
+	if (job_info->rc) {
+		g_object_unref (G_OBJECT (job_info->rc));
+	}
+	job_info->rc = g_object_ref (job_render->rc);
+
+	if (job_render->include_links) {
+		if (job_info->link_mapping)
+			ev_link_mapping_free (job_info->link_mapping);
 		job_info->link_mapping = job_render->link_mapping;
-	if (job_render->text_mapping)
-		job_info->text_mapping = job_render->text_mapping;	
+	}
+
+	if (job_render->include_images) {
+		if (job_info->image_mapping)
+			ev_image_mapping_free (job_info->image_mapping);
+		job_info->image_mapping = job_render->image_mapping;
+	}
+
+	if (job_render->include_text) {
+		if (job_info->text_mapping)
+			gdk_region_destroy (job_info->text_mapping);
+		job_info->text_mapping = job_render->text_mapping;
+	}
+
+	if (job_render->include_selection) {
+		if (job_info->selection) {
+			cairo_surface_destroy (job_info->selection);
+			job_info->selection = NULL;
+		}
+		if (job_info->selection_region) {
+			gdk_region_destroy (job_info->selection_region);
+			job_info->selection_region = NULL;
+		}
+		
+		job_info->selection_points = job_render->selection_points;
+		job_info->selection_region = gdk_region_copy (job_render->selection_region);
+		job_info->selection = cairo_surface_reference (job_render->selection);
+		g_assert (job_info->selection_points.x1 >= 0);
+		job_info->points_set = TRUE;
+	}
+
+	if (job_info->job) {
+		g_signal_handlers_disconnect_by_func (job_info->job,
+						      G_CALLBACK (job_finished_cb),
+						      pixbuf_cache);
+		ev_job_queue_remove_job (job_info->job);
+		g_object_unref (G_OBJECT (job_info->job));
+		job_info->job = NULL;
+	}
 }
 
 static CacheJobInfo *
@@ -476,12 +507,12 @@ ev_pixbuf_cache_clear_job_sizes (EvPixbufCache *pixbuf_cache,
 	page_cache = ev_page_cache_get (pixbuf_cache->document);
 
 	for (i = 0; i < PAGE_CACHE_LEN (pixbuf_cache); i++) {
-		check_job_size_and_unref (pixbuf_cache->job_list + i, page_cache, scale);
+		check_job_size_and_unref (pixbuf_cache, pixbuf_cache->job_list + i, page_cache, scale);
 	}
 
 	for (i = 0; i < pixbuf_cache->preload_cache_size; i++) {
-		check_job_size_and_unref (pixbuf_cache->prev_job + i, page_cache, scale);
-		check_job_size_and_unref (pixbuf_cache->next_job + i, page_cache, scale);
+		check_job_size_and_unref (pixbuf_cache, pixbuf_cache->prev_job + i, page_cache, scale);
+		check_job_size_and_unref (pixbuf_cache, pixbuf_cache->next_job + i, page_cache, scale);
 	}
 }
 
@@ -512,6 +543,7 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 	gboolean include_links = FALSE;
 	gboolean include_text = FALSE;
 	gboolean include_selection = FALSE;
+	gboolean include_images = FALSE;
 	int width, height;
 	GdkColor *text, *base;
 
@@ -521,9 +553,9 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 	ev_page_cache_get_size (page_cache, page, rotation,
 				scale, &width, &height);
 
-	if (job_info->pixbuf &&
-	    gdk_pixbuf_get_width (job_info->pixbuf) == width &&
-	    gdk_pixbuf_get_height (job_info->pixbuf) == height)
+	if (job_info->surface &&
+	    cairo_image_surface_get_width (job_info->surface) == width &&
+	    cairo_image_surface_get_height (job_info->surface) == height)
 		return;
 
 	/* make a new job now */
@@ -538,9 +570,11 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 	/* Figure out what else we need for this job */
 	if (job_info->link_mapping == NULL)
 		include_links = TRUE;
+	if (job_info->image_mapping == NULL)
+		include_images = TRUE;
 	if (job_info->text_mapping == NULL)
 		include_text = TRUE;
-	if (new_selection_pixbuf_needed (pixbuf_cache, job_info, page, scale)) {
+	if (new_selection_surface_needed (pixbuf_cache, job_info, page, scale)) {
 		include_selection = TRUE;
 	}
 
@@ -551,9 +585,10 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 	job_info->job = ev_job_render_new (pixbuf_cache->document,
 					   job_info->rc,
 					   width, height,
-					   &(job_info->new_points),
+					   &(job_info->target_points),
 					   text, base,
 					   include_links,
+					   include_images,
 					   include_text,
 					   include_selection);
 	ev_job_queue_add_job (job_info->job, priority);
@@ -636,9 +671,9 @@ ev_pixbuf_cache_set_page_range (EvPixbufCache  *pixbuf_cache,
 	ev_pixbuf_cache_add_jobs_if_needed (pixbuf_cache, rotation, scale);
 }
 
-GdkPixbuf *
-ev_pixbuf_cache_get_pixbuf (EvPixbufCache *pixbuf_cache,
-			    gint           page)
+cairo_surface_t *
+ev_pixbuf_cache_get_surface (EvPixbufCache *pixbuf_cache,
+			     gint           page)
 {
 	CacheJobInfo *job_info;
 
@@ -650,9 +685,10 @@ ev_pixbuf_cache_get_pixbuf (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 	}
 
-	return job_info->pixbuf;
+	return job_info->surface;
 }
 
 GList *
@@ -669,32 +705,62 @@ ev_pixbuf_cache_get_link_mapping (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 	}
-	
+
 	return job_info->link_mapping;
 }
 
+GList *
+ev_pixbuf_cache_get_image_mapping (EvPixbufCache *pixbuf_cache,
+				   gint           page)
+{
+	CacheJobInfo *job_info;
+
+	if (!EV_IS_DOCUMENT_IMAGES (pixbuf_cache->document))
+		return NULL;
+	
+	job_info = find_job_cache (pixbuf_cache, page);
+	if (job_info == NULL)
+		return NULL;
+
+	/* We don't need to wait for the idle to handle the callback */
+	if (job_info->job &&
+	    EV_JOB (job_info->job)->finished) {
+		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
+	}
+
+	return job_info->image_mapping;
+}
+
 static gboolean
-new_selection_pixbuf_needed (EvPixbufCache *pixbuf_cache,
+new_selection_surface_needed (EvPixbufCache *pixbuf_cache,
 			     CacheJobInfo  *job_info,
 			     gint           page,
 			     gfloat         scale)
 {
 	EvPageCache *page_cache;
-	gint width, height;
 
 	if (job_info->selection) {
-		page_cache = ev_page_cache_get (pixbuf_cache->document);
-		ev_page_cache_get_size (page_cache, page, job_info->rc->rotation,
-					scale, &width, &height);
+		gint width, height;
+		gint selection_width, selection_height;
 		
-		if (width != gdk_pixbuf_get_width (job_info->selection) ||
-		    height != gdk_pixbuf_get_height (job_info->selection))
+		page_cache = ev_page_cache_get (pixbuf_cache->document);
+		ev_page_cache_get_size (page_cache, page,
+					job_info->rc->rotation,
+					scale, &width, &height);
+
+		selection_width = cairo_image_surface_get_width (job_info->selection);
+		selection_height = cairo_image_surface_get_height (job_info->selection);
+		
+		if (width != selection_width || height != selection_height)
 			return TRUE;
 	} else {
-		if (job_info->new_points.x1 >= 0)
+		if (job_info->points_set)
 			return TRUE;
 	}
+	
 	return FALSE;
 }
 
@@ -704,17 +770,17 @@ clear_selection_if_needed (EvPixbufCache *pixbuf_cache,
 			   gint           page,
 			   gfloat         scale)
 {
-	if (new_selection_pixbuf_needed (pixbuf_cache, job_info, page, scale)) {
+	if (new_selection_surface_needed (pixbuf_cache, job_info, page, scale)) {
 		if (job_info->selection)
-			g_object_unref (job_info->selection);
+			cairo_surface_destroy (job_info->selection);
 		job_info->selection = NULL;
 		job_info->selection_points.x1 = -1;
 	}
 }
 
 GdkRegion *
-ev_pixbuf_cache_get_text_mapping      (EvPixbufCache *pixbuf_cache,
-				       gint           page)
+ev_pixbuf_cache_get_text_mapping (EvPixbufCache *pixbuf_cache,
+				  gint           page)
 {
 	CacheJobInfo *job_info;
 
@@ -726,6 +792,7 @@ ev_pixbuf_cache_get_text_mapping      (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 	}
 	
 	return job_info->text_mapping;
@@ -760,13 +827,13 @@ ev_pixbuf_cache_style_changed (EvPixbufCache *pixbuf_cache)
 
 		job_info = pixbuf_cache->prev_job + i;
 		if (job_info->selection) {
-			g_object_unref (G_OBJECT (job_info->selection));
+			cairo_surface_destroy (job_info->selection);
 			job_info->selection = NULL;
 		}
 
 		job_info = pixbuf_cache->next_job + i;
 		if (job_info->selection) {
-			g_object_unref (G_OBJECT (job_info->selection));
+			cairo_surface_destroy (job_info->selection);
 			job_info->selection = NULL;
 		}
 	}
@@ -776,17 +843,17 @@ ev_pixbuf_cache_style_changed (EvPixbufCache *pixbuf_cache)
 
 		job_info = pixbuf_cache->job_list + i;
 		if (job_info->selection) {
-			g_object_unref (G_OBJECT (job_info->selection));
+			cairo_surface_destroy (job_info->selection);
 			job_info->selection = NULL;
 		}
 	}
 }
 
-GdkPixbuf *
-ev_pixbuf_cache_get_selection_pixbuf (EvPixbufCache  *pixbuf_cache,
-				      gint            page,
-				      gfloat          scale,
-				      GdkRegion     **region)
+cairo_surface_t *
+ev_pixbuf_cache_get_selection_surface (EvPixbufCache  *pixbuf_cache,
+				       gint            page,
+				       gfloat          scale,
+				       GdkRegion     **region)
 {
 	CacheJobInfo *job_info;
 
@@ -799,7 +866,7 @@ ev_pixbuf_cache_get_selection_pixbuf (EvPixbufCache  *pixbuf_cache,
 		return NULL;
 
 	/* No selection on this page */
-	if (job_info->new_points.x1 < 0)
+	if (!job_info->points_set)
 		return NULL;
 
 	/* Update the rc */
@@ -821,7 +888,7 @@ ev_pixbuf_cache_get_selection_pixbuf (EvPixbufCache  *pixbuf_cache,
 	 * _should_ be able to get rid of the doc_mutex, so the synchronicity
 	 * doesn't kill us.  Rendering a few glyphs should really be fast.
 	 */
-	if (ev_rect_cmp (&(job_info->new_points), &(job_info->selection_points))) {
+	if (ev_rect_cmp (&(job_info->target_points), &(job_info->selection_points))) {
 		EvRectangle *old_points;
 		GdkColor *text, *base;
 
@@ -840,7 +907,7 @@ ev_pixbuf_cache_get_selection_pixbuf (EvPixbufCache  *pixbuf_cache,
 		job_info->selection_region =
 			ev_selection_get_selection_region (EV_SELECTION (pixbuf_cache->document),
 							   job_info->rc,
-							   &(job_info->new_points));
+							   &(job_info->target_points));
 
 		gtk_widget_ensure_style (pixbuf_cache->view);
 
@@ -848,10 +915,10 @@ ev_pixbuf_cache_get_selection_pixbuf (EvPixbufCache  *pixbuf_cache,
 
 		ev_selection_render_selection (EV_SELECTION (pixbuf_cache->document),
 					       job_info->rc, &(job_info->selection),
-					       &(job_info->new_points),
+					       &(job_info->target_points),
 					       old_points,
 					       text, base);
-		job_info->selection_points = job_info->new_points;
+		job_info->selection_points = job_info->target_points;
 		ev_document_doc_mutex_unlock ();
 	}
 	if (region)
@@ -863,19 +930,18 @@ static void
 update_job_selection (CacheJobInfo    *job_info,
 		      EvViewSelection *selection)
 {
-	if (job_info->selection == NULL)
-		job_info->selection_points.x1 = -1;
-	job_info->new_points = selection->rect;
+	job_info->points_set = TRUE;		
+	job_info->target_points = selection->rect;
 }
 
 static void
 clear_job_selection (CacheJobInfo *job_info)
 {
+	job_info->points_set = FALSE;
 	job_info->selection_points.x1 = -1;
-	job_info->new_points.x1 = -1;
 
 	if (job_info->selection) {
-		g_object_unref (job_info->selection);
+		cairo_surface_destroy (job_info->selection);
 		job_info->selection = NULL;
 	}
 }
@@ -896,6 +962,9 @@ ev_pixbuf_cache_set_selection_list (EvPixbufCache *pixbuf_cache,
 
 	g_return_if_fail (EV_IS_PIXBUF_CACHE (pixbuf_cache));
 
+	if (!EV_IS_SELECTION (pixbuf_cache->document))
+		return;
+
 	page_cache = ev_page_cache_get (pixbuf_cache->document);
 
 	/* We check each area to see what needs updating, and what needs freeing; */
