@@ -23,21 +23,29 @@
 
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <gtk/gtk.h>
+#ifdef WITH_GNOME_PRINT
+#include <libgnomeprintui/gnome-print-dialog.h>
+#endif
 
 G_BEGIN_DECLS
 
-GdkPixbuf *ev_pixbuf_add_shadow (GdkPixbuf *src, int size,
-				 int x_offset, int y_offset, double opacity);
+GdkPixbuf*		ev_pixbuf_add_shadow (GdkPixbuf *src, int size,
+					      int x_offset, int y_offset, double opacity);
 
-void       ev_print_region_contents (GdkRegion *region);
+void			ev_print_region_contents (GdkRegion *region);
 
-#ifndef HAVE_G_FILE_SET_CONTENTS
+void ev_gui_menu_position_tree_selection (GtkMenu   *menu,
+					  gint      *x,
+					  gint      *y,
+					  gboolean  *push_in,
+					  gpointer   user_data);
 
-gboolean   ev_file_set_contents (const gchar *filename,
-		    		 const gchar *contents,
-		    		 gssize	     length,
-		    		 GError	   **error);
-#endif /* HAVE_G_FILE_SET_CONTENTS */
+#ifdef WITH_GNOME_PRINT
+GnomePrintConfig* 	load_print_config_from_file (void);
+void       		save_print_config_to_file (GnomePrintConfig *config);
+gboolean		using_postscript_printer (GnomePrintConfig *config);
+gboolean		using_pdf_printer (GnomePrintConfig *config);
+#endif
 
 G_END_DECLS
 
